﻿using Microsoft.Win32;
using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;


namespace ClipboardViewer
{
    public partial class Form1 : Form
    {
        
        // Clipboard stuff
        [DllImport("User32.dll")]
        public static extern IntPtr SetClipboardViewer(IntPtr hWndNewViewer);

        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        public static extern bool ChangeClipboardChain(IntPtr hWndRemove, IntPtr hWndNewNext);

        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        public static extern int SendMessage(IntPtr hwnd, int wMsg, IntPtr wParam, IntPtr lParam);

        IntPtr nextClipboardViewer;

        // tried to fix the window activating itself the first time it is set to topmost while not active
        //  which shouldn't happen anyway and only happens the first time
        //  but it doesn't work with this anyway
        //protected override bool ShowWithoutActivation => true;//base.ShowWithoutActivation;

        /*
         * ContentTypes: 00000001 = Text
         *               00000010 = Unicode
         *               ...
         *               00001000 = Bitmap
         *               ...
         */
        const byte ct_TEXT = 1;
        const byte ct_UNICODE = 2;
        const byte ct_RTF = 4;
        const byte ct_BITMAP = 8;
        const byte ct_AUDIO = 16;
        const byte ct_FILE = 32;
        byte contentTypes = 0;

        // used to prevent window from being sent behind other windows after program start
        bool bfirstStart = true;
        // indicate if load procedure is done, including reading settings from clipboard.
        // used to prevent reading clipboard (with its consequences) before that.
        bool bProgramStarted = false;

        // Registry stuff
        bool bRegistryRead = false;
        RegistryKey baseRKey;
        const String SOFTWARE = "SOFTWARE";
        // settings names:
        const String APPNAME = "AppName";
        const String VERSION = "Version";
        const String BRING_TO_FRONT = "BringToFront";
        const String REREGISTER = "Reregister";
        const String FORCE_GARBAGE_COLLECTION = "ForceGarbageCollection";
        const String SHOW_EXCEPTIONS = "ShowExceptions";
        const String DX = "dX";
        const String DY = "dY";
        const String X = "X";
        const String Y = "Y";

        // used to position elements
        int num_elements = 0;
        // option if content should be displayed or not
        bool bHideContent = false;
        // proxy visibility vars for the various elements
        bool bTextBoxVisible = false;
        bool bRichTextBoxVisible = false;
        bool bPictureBoxVisible = false;
        // secret hidden setting
        bool bForceGarbageCollection = false;
        // how exceptions shall be shown
        enum ShowExceptionModes
        {
            None = 0,
            StatusBar = 1,
            ImmediateFull = 2
        };
        ShowExceptionModes ShowExceptionMode = ShowExceptionModes.StatusBar;
        Exception lastException = null;

        int iNumberOfFiledropElements = 0;
        //int iBitmapWidth = 0;
        //int iBitmapHeigth = 0;

        // default number of retries when checking clipboard
        const int CLIPBOARD_CHECK_RETRIES = 3;

        public Form1()
        {
            InitializeComponent();
        }

        private void readRegistry()
        {
            // Registry: check for stored values. if none are found, write defaults
            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            if (this.baseRKey == null)
            {
                _log("baseRKey == null, writing defaults");
                // defaults
                this.BringToFrontOption.Checked = true;
                this.registerForClipboardChangesToolStripMenuItem.Checked = true;
                this.setShowExceptionMode(ShowExceptionModes.StatusBar, false);

                this.writeRegistryDefaults();
            }
            else
            {
                // load values from registry
                int itmp;
                String stmp;

                itmp = (int)this.baseRKey.GetValue(BRING_TO_FRONT, 3);
                _log("BRING TO FRONT = " + itmp);
                if (itmp >= 1)
                {
                    this.BringToFrontOption.Checked = true;
                }
                else
                {
                    this.BringToFrontOption.Checked = false;
                }

                itmp = (int)this.baseRKey.GetValue(FORCE_GARBAGE_COLLECTION, -1);
                _log("FORCE GARBAGE COLLECTION = " + itmp);
                if (itmp >= 1)
                {
                    this.bForceGarbageCollection = true;
                }
                else
                {
                    this.bForceGarbageCollection = false;
                }
                
                itmp = (int)this.baseRKey.GetValue(SHOW_EXCEPTIONS, 1);
                _log("SHOW EXCEPTIONS = " + itmp);
                if (itmp == 0)
                {
                    this.setShowExceptionMode(ShowExceptionModes.None, false);
                }
                else if (itmp == 1)
                {
                    this.setShowExceptionMode(ShowExceptionModes.StatusBar, false);
                }
                else if (itmp == 2)
                {
                    this.setShowExceptionMode(ShowExceptionModes.ImmediateFull, false);
                }

                itmp = (int)this.baseRKey.GetValue(REREGISTER, 3);
                _log("REREGISTER = " + itmp);
                if (itmp >= 1)
                {
                    this.registerForClipboardChangesToolStripMenuItem.Checked = true;
                }
                else
                {
                    this.registerForClipboardChangesToolStripMenuItem.Checked = false;
                }

                // window position and size
                // don't write this to registry by default
                itmp = (int)this.baseRKey.GetValue(DX, -1);
                if (itmp > 0)
                {
                    this.Width = itmp;
                }
                itmp = (int)this.baseRKey.GetValue(DY, -1);
                if (itmp > 0)
                {
                    this.Height = itmp;
                }
                Point pWindowLocation = new Point((int)this.baseRKey.GetValue(X, -1), (int)this.baseRKey.GetValue(Y, -1));
                _log("Loaded: X = " + pWindowLocation.X + "; Y = " + pWindowLocation.Y + ";");

                // check if window is outside visible area
                // for example if window was on secondary monitor that's no longer present
                Rectangle rWorkingArea = Screen.GetWorkingArea(pWindowLocation);
                _log("Working Area: Top = " + rWorkingArea.Top + "; Bottom = " + rWorkingArea.Bottom + "; Left = " + rWorkingArea.Left + "; Right = " + rWorkingArea.Right + ";");
                if (pWindowLocation.Y < rWorkingArea.Top) pWindowLocation.Y = rWorkingArea.Top;
                if (pWindowLocation.Y > rWorkingArea.Bottom) pWindowLocation.Y = rWorkingArea.Bottom - this.Height;
                if (pWindowLocation.X < rWorkingArea.Left) pWindowLocation.X = rWorkingArea.Left;
                if (pWindowLocation.X > rWorkingArea.Right) pWindowLocation.X = rWorkingArea.Right - this.Width;

                _log("Adjusted: X = " + pWindowLocation.X + "; Y = " + pWindowLocation.Y + ";");

                this.SetDesktopLocation(pWindowLocation.X, pWindowLocation.Y);

                // check if version matches
                // note: if variables change, this check might be neccessary *bevore* loading values
                stmp = (String)this.baseRKey.GetValue(VERSION, "");
                _log("VERSION = " + stmp);
                if (stmp != this.ProductVersion)
                {
                    // The ClipboardViewer was updated!
                    this.writeRegistryUpdates();
                }
                this.baseRKey.Close();
            }
            bRegistryRead = true;
        }

        private void _log(String text)
        {
            this.logBox.Font = SystemFonts.DefaultFont;
            // the logBox is invisible by default and was only created to see a log when first checking
            // registry writing/reading functionality. might be removed any time.
            this.logBox.Text = text + "\r\n" + this.logBox.Text;
        }

        private void writeRegistryDefaults()
        {
            // first time program starts -> write defaults
            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE, true);
            this.baseRKey.CreateSubKey(this.ProductName);
            this.baseRKey.Close();

            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            //_log("baseRKey should be in SOFTWARE\\" + this.ProductName + ", is " + this.baseRKey);
            this.baseRKey.SetValue(APPNAME, this.ProductName);
            this.baseRKey.SetValue(VERSION, this.ProductVersion);
            this.baseRKey.SetValue(BRING_TO_FRONT, 1);
            this.baseRKey.SetValue(REREGISTER, 1);
            this.baseRKey.SetValue(FORCE_GARBAGE_COLLECTION, 0);
            this.writeShowExceptionModeToRegistry(this.baseRKey);
            this.baseRKey.Close();
        }
        private void writeRegistryUpdates()
        {
            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            this.baseRKey.SetValue(VERSION, this.ProductVersion);
            // instead of writing defaults, use previously loaded values
            if (this.BringToFrontOption.Checked) this.baseRKey.SetValue(BRING_TO_FRONT, 1);
            if (!this.BringToFrontOption.Checked) this.baseRKey.SetValue(BRING_TO_FRONT, 0);
            if (this.registerForClipboardChangesToolStripMenuItem.Checked) this.baseRKey.SetValue(REREGISTER, 1);
            if (!this.registerForClipboardChangesToolStripMenuItem.Checked) this.baseRKey.SetValue(REREGISTER, 0);
            if (this.bForceGarbageCollection) this.baseRKey.SetValue(FORCE_GARBAGE_COLLECTION, 1);
            if (!this.bForceGarbageCollection) this.baseRKey.SetValue(FORCE_GARBAGE_COLLECTION, 0);
            this.writeShowExceptionModeToRegistry(this.baseRKey);
            
            this.baseRKey.Close();
        }

        private void writeShowExceptionModeToRegistry(RegistryKey rkey)
        {
            bool ownRKey = false;
            if (rkey == null)
            {
                ownRKey = true;
                rkey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            }
            int sem = 1;
            switch (this.ShowExceptionMode)
            {
                case ShowExceptionModes.None:
                    sem = 0;
                    break;
                case ShowExceptionModes.StatusBar:
                    sem = 1;
                    break;
                case ShowExceptionModes.ImmediateFull:
                    sem = 2;
                    break;
            }
            rkey.SetValue(SHOW_EXCEPTIONS, sem);
            if (ownRKey)
            {
                rkey.Close();
            }
        }

        private void reRegisterHandler()
        {
            // deactivate bring to front, otherwise ClipboardViewer pops up without a clipboard change
            // TODO: use a seperate thread to do this?
            bool tmp = this.BringToFrontOption.Checked;
            this.BringToFrontOption.Checked = false;

            // first deregister
            ChangeClipboardChain(this.Handle, nextClipboardViewer);

            // then reregister
            nextClipboardViewer = SetClipboardViewer(this.Handle);

            this.BringToFrontOption.Checked = tmp;
        }

        protected override void WndProc(ref Message m)
        {
            const int WM_DRAWCLIPBOARD = 0x308;
            const int WM_CHANGEBCHAIN = 0x030D;

            switch (m.Msg)
            {
                case WM_DRAWCLIPBOARD:
                    // new content in clipboard
                    // display and send to next app
                    SendMessage(nextClipboardViewer, m.Msg, m.WParam, m.LParam);
                    this.checkClipboard(CLIPBOARD_CHECK_RETRIES);
                    break;

                case WM_CHANGEBCHAIN:
                    if (m.WParam == nextClipboardViewer)
                        // someone leaves the clipboard chain
                        // is it the program after ClipboardViewer? then delete it
                        nextClipboardViewer = m.LParam;
                    else
                        // else pass message to the program after CV
                        SendMessage(nextClipboardViewer, m.Msg, m.WParam, m.LParam);
                    break;

                default:
                    base.WndProc(ref m);
                    break;
            }
        }

        private void checkClipboard(int retries)
        {
            // prevent reading of clipboard before settings have been loaded
            if (!this.bProgramStarted)
            {
                return;
            }

            this.contentTypes = 0;
            this.num_elements = 0;

            // clear some memory and reset all fonts
            //  unfortunately if ClipboardViewer starts with windows, the wrong font is used, so manually
            //  set it to system default
            this.textBox.Font = SystemFonts.DefaultFont;
            this.textBox.Clear();
            //this.textBox.Text = "";
            this.richTextBox.Font = SystemFonts.DefaultFont;
            this.richTextBox.Clear();
            //this.richTextBox.Text = "";
            this.pictureBox.Image = null;

            // forget previous styling (i.e. Font, Text Color, ...)
            // hopefully .Clear() already does that
            // also .ResetText() doesn't work reliably...
            //this.richTextBox.ResetText();

            if (this.bForceGarbageCollection) {
                System.GC.Collect();
            };

            try
            {
                IDataObject idata = Clipboard.GetDataObject();

                // first check for standard text, then unicode text, then RichText
                // up to now I haven't seen these text types simultaneously, so reuse richTextBox for all of them
                if (idata.GetDataPresent(DataFormats.Text))
                {
                    this.contentTypes |= ct_TEXT;
                    try
                    {
                        this.richTextBox.Text = (String)idata.GetData(DataFormats.Text);
                    }
                    catch
                    {
                        // TODO: Error reporting. For example to console. Or whatever?
                    }
                }

                if (idata.GetDataPresent(DataFormats.UnicodeText))
                {
                    this.contentTypes |= ct_UNICODE;
                    try
                    {
                        this.richTextBox.Text = (String)idata.GetData(DataFormats.UnicodeText);
                    }
                    catch
                    {
                        // TODO: Error reporting. For example to console. Or whatever?
                    }
                }

                if (idata.GetDataPresent(DataFormats.Rtf))
                {
                    this.contentTypes |= ct_RTF;
                    try
                    {
                        this.richTextBox.Rtf = (String)idata.GetData(DataFormats.Rtf);
                    }
                    catch
                    {
                        // TODO: Error reporting. For example to console. Or whatever?
                    }
                }

                // check for image
                if (idata.GetDataPresent(DataFormats.Bitmap))
                {
                    this.contentTypes |= ct_BITMAP;
                    try
                    {
                        this.pictureBox.Image = (Bitmap)idata.GetData(DataFormats.Bitmap);
                    }
                    catch
                    {
                        // TODO: Error reporting. For example to console. Or whatever?
                    }
                }

                // check for audio (display icon/picture and/or text to indicate presence of audio)
                if (idata.GetDataPresent(DataFormats.WaveAudio))
                {
                    this.contentTypes |= ct_AUDIO;
                }

                // check for FileDrop
                // as all text goes to richTextBox, use "normal" textBox to display Filenames
                // TODO: display icons?
                if (idata.GetDataPresent(DataFormats.FileDrop))
                {
                    this.contentTypes |= ct_FILE;

                    System.Collections.Specialized.StringCollection sc = Clipboard.GetFileDropList();
                    iNumberOfFiledropElements = sc.Count;
                    // convert filedroplist to string
                    string[] shing = new string[sc.Count];
                    sc.CopyTo(shing, 0);
                    /*
                     * Display filelist in textBox, so richTextBox is "free" for actual strings (might coexist,
                     * try copying from foobar2000)
                     */
                    this.textBox.Text = string.Join("\r\n", shing);
                }

                // TODO: make display of length/type optional; text: also display lines; display memory usage of clipboard? format large numbers (2100 -> either 2,100 or 2.1k)
                if (contentTypes == 0)
                {
                    // no (known) content in clipboard
                    this.bTextBoxVisible = false;
                    this.bRichTextBoxVisible = false;
                    this.bPictureBoxVisible = false;
                    this.typeLabel.Text = "Unknown type or empty clipboard";
                }
                else
                {
                    this.typeLabel.Text = "";
                    if ((contentTypes & ct_TEXT) > 0 || (contentTypes & ct_UNICODE) > 0 || (contentTypes & ct_RTF) > 0)
                    {
                        // if *any* text should be displayed, make richtextbox visible
                        this.bRichTextBoxVisible = true;
                        this.num_elements += 1;

                        String texttype = "";
                        int chars = 0;
                        int lines = 0;
                        int non_empty_lines = 0;

                        // if multiple text types are available, only display an indicator for fancyest
                        if ((contentTypes & ct_RTF) > 0)
                        {
                            texttype = "RTF";
                        }
                        else
                        {
                            if ((contentTypes & ct_UNICODE) > 0)
                            {
                                texttype = "UnicodeText";
                            }
                            else
                            {
                                texttype = "Text";
                            }
                        }
                        // count chars / lines
                        if (texttype != "") {
                            chars = this.richTextBox.TextLength;
                            String[] ll = this.richTextBox.Text.Split(Convert.ToChar("\n"));
                            lines = ll.Length;
                            for (int i = 0; i < ll.Length; i++)
                            {
                                if (ll[i].Trim().Length > 0)
                                {
                                    non_empty_lines++;
                                }
                            }
                            this.typeLabel.Text += texttype + "(" + ((lines > 1) ? ("C" + Convert.ToString(chars) + "L" + Convert.ToString(lines) + "N" + Convert.ToString(non_empty_lines)) : (Convert.ToString(chars)))+")";
                        }
                    }
                    else
                    {
                        // no text, no richTextBox
                        this.bRichTextBoxVisible = false;
                    }

                    // display textBox if FileDrop in Clipboard
                    if ((contentTypes & ct_FILE) > 0)
                    {
                        this.bTextBoxVisible = true;
                        this.num_elements += 1;
                        this.typeLabel.Text += "FileDrop(" + Convert.ToString(iNumberOfFiledropElements) + ") ";
                    }
                    else
                    {
                        this.bTextBoxVisible = false;
                    }

                    // display pictureBox if picture in clipboard
                    if ((contentTypes & ct_BITMAP) > 0)
                    {
                        this.bPictureBoxVisible = true;
                        this.num_elements += 1;
                        this.typeLabel.Text += "Bitmap(" + Convert.ToString(this.pictureBox.Image.Width) + "x" + Convert.ToString(this.pictureBox.Image.Height) + ") ";
                    }
                    else
                    {
                        this.bPictureBoxVisible = false;
                    }

                    // only display an indicator for audio
                    if ((contentTypes & ct_AUDIO) > 0)
                    {
                        this.typeLabel.Text += "Audio ";
                    }

                    // do the layouting
                    this.applySizes();
                    // do the visibility
                    this.applyVisibility();

                    if (this.BringToFrontOption.Checked && !this.bHideContent)
                    {
                        // reset the timer that sends window to back, so the new clipboard content is visible some time
                        this.DisableTopMost.Stop();

                        // bring to front and enable the timer that sends the window back again
                        // calling this.Show() before setting it to TopMost prevents the form of stealing
                        //  focus the first time it is set to topmost while not already having focus.
                        //  don't ask why.
                        // using .Show() alone doesn't bring it to front, it stays behind other windows
                        this.Show();
                        this.TopMost = true;
                        this.DisableTopMost.Start();
                    }
                }
            }
            catch (ExternalException e)
            {
                // this exception is, for example, raised when the current clipboard content is from MS Word and Word is closed m(
                lastException = e;

                // try again if retry-counter not 0
                if (retries > 0)
                {
                    // indicate that an exception has occured
                    if (this.ShowExceptionMode != ShowExceptionModes.None)
                    {
                        this.typeLabel.Text = "ExternalException occured (retrying...)";
                    }
                    // TODO: wait a moment (~ a second)?
                    // warning: just calling Thread.Sleep can apparentliy lead to lengthy unresponiveness
                    //  I guess that checkClipboard() can be called by a clipboard change while the thread
                    //  is still sleeping
                    this.checkClipboard(retries - 1);
                }
                else
                {
                    if (this.ShowExceptionMode == ShowExceptionModes.ImmediateFull)
                    {
                        MessageBox.Show(e.ToString());
                    }
                    else if (this.ShowExceptionMode == ShowExceptionModes.StatusBar)
                    {
                        this.typeLabel.Text = "ExternalException occured";
                    }
                    // else ShowExceptionMode == ShowExceptionModes.None
                }
            }
            catch (Exception e)
            {
                lastException = e;
                // A wild error appears!
                if (this.ShowExceptionMode == ShowExceptionModes.ImmediateFull)
                {
                    MessageBox.Show(e.ToString());
                }
                else if (this.ShowExceptionMode == ShowExceptionModes.StatusBar)
                {
                    this.typeLabel.Text = "Exception occured";
                }
                // else ShowExceptionMode == ShowExceptionModes.None
            }
        }

        private void applyVisibility()
        {
            // if everything shall be hidden, well, hide everything, but display an indicator
            // else display what shall be displayed (i.e. that which has content, see CheckClipboard())
            if (this.bHideContent)
            {
                this.textBox.Visible = false;
                this.richTextBox.Visible = false;
                this.pictureBox.Visible = false;
                this.toolStripStatusVisibilityLabel.Visible = true;
            } else {
                this.textBox.Visible = this.bTextBoxVisible;
                this.richTextBox.Visible = this.bRichTextBoxVisible;
                this.pictureBox.Visible = this.bPictureBoxVisible;
                this.toolStripStatusVisibilityLabel.Visible = false;
            }
        }

        private void applySizes()
        {
            /*
             * All displaying objects should touch the visible area of the window from the inside.
             * 
             * If more than one object should be visible (num_elements > 1), arrange them somehow.
             * 
             * num_elements is set when clipboard content changes and is a global variable. As applySizes is also
             * called from other functions, it is difficult to make it into an argument to the function.
             */

            this.pictureBox.Left = 0;
            this.textBox.Left = 0;
            this.richTextBox.Left = 0;
            this.pictureBox.Top = 0;
            this.textBox.Top = -1;
            this.richTextBox.Top = 0;

            // TODO: Magic Number 16. Find a better way.
            this.pictureBox.Width = this.Width-16;
            this.textBox.Width = this.Width-16;
            this.richTextBox.Width = this.Width - 16;
            this.pictureBox.Height = this.statusStrip1.Top;
            this.textBox.Height = this.statusStrip1.Top;
            this.richTextBox.Height = this.statusStrip1.Top;

            if (num_elements == 2)
            {
                if (this.bRichTextBoxVisible == true)
                {
                    this.richTextBox.Height /= 2;
                    if (this.bPictureBoxVisible == true)
                    {
                        this.pictureBox.Height /= 2;
                        this.pictureBox.Top += this.richTextBox.Height;
                    }
                    else
                    {
                        this.textBox.Height /= 2;
                        this.textBox.Top += this.richTextBox.Height;
                    }
                }
                else
                {
                    this.pictureBox.Height /= 2;
                    this.textBox.Height /= 2;
                    this.textBox.Top += this.richTextBox.Height;
                }
            }
            else if (num_elements == 3)
            {
                this.richTextBox.Height /= 3;
                this.pictureBox.Height /= 3;
                this.textBox.Height /= 3;
                this.pictureBox.Top += this.richTextBox.Height;
                this.textBox.Top += this.richTextBox.Height;
                this.textBox.Top += this.pictureBox.Height;
            }
        }

        private void Form1_ResizeEnd(object sender, EventArgs e)
        {
            if (this.WindowState == FormWindowState.Normal)
            {
                // just to be sure, only wrize size to registry if normal window state
                this.applySizes();
                this.writeSizeToRegistry();
            };
        }

        private void writeSizeToRegistry()
        {
            if (bRegistryRead) // make sure not to overwrite the data before reading it first
            {
                this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
                //_log("baseRKey should be in SOFTWARE\\" + this.ProductName + ", is " + this.baseRKey);
                this.baseRKey.SetValue(DX, this.Width);
                this.baseRKey.SetValue(DY, this.Height);
                this.baseRKey.Close();
            }
        }

        private void writeWindowPositionToRegistry()
        {
            if (bRegistryRead) // make sure not to overwrite the data before reading it first
            {
                this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
                //_log("baseRKey should be in SOFTWARE\\" + this.ProductName + ", is " + this.baseRKey);
                this.baseRKey.SetValue(X, this.Left);
                this.baseRKey.SetValue(Y, this.Top);
                this.baseRKey.Close();
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            this.reRegisterTimer.Interval = 5 * 60 * 1000; // should be 5 minutes
            this.nextClipboardViewer = SetClipboardViewer(this.Handle);

            this.readRegistry();

            this.reRegisterTimer.Enabled = this.registerForClipboardChangesToolStripMenuItem.Checked;

            this.applySizes();
            // just in case
            this.applyVisibility();

            // enable debug output
            string[] args = System.Environment.GetCommandLineArgs();
            if (Array.IndexOf(args, "-d") >= 0)
            {
                this.logBox.Visible = true;
            }
            else
            {
                this.logBox.Visible = false;
            }

            this.bProgramStarted = true;
            this.checkClipboard(CLIPBOARD_CHECK_RETRIES);
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show(this.ProductName + " v" + this.ProductVersion + "\r\nby Leo Hackstein 2011-2019\r\nhttps://oneko.de/prog/ClipboardViewer.html\r\n\r\nDisplays the current content of the Windows Clipboard.", "About", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            /*
             * Make sure to leave the Clipboard Chain properly, otherwise programs behind ClipboardViewer don't get notified
             * on clipboard changes any more.
             */
            ChangeClipboardChain(this.Handle, nextClipboardViewer);
        }

        private void reRegisterTimer_Tick(object sender, EventArgs e)
        {
            /*
             * Some programs destroy the clipboard chain (for example Synergy http://synergy-foss.org/ (tested ~2011)), so all programs
             * that are already in the chain tend to break. Reregistering moves the program to the start of the chain and
             * so in front of the misbehaving programs.
             * 
             * Also, sometimes ClipboardViewer doesn't get notified aboud clipboard changes any more.
             * I have no idea why, but reregistering makes it function again, so I blame the clipboard chain, Windows or other
             * misbehaving programs.
             */
            reRegisterHandler();
        }

        private void registerForClipboardChangesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*
             * Instead of instantly reregistering, this has been converted to an option whether or not to automatically
             * reregister for clipboard changes.
             */

            this.reRegisterTimer.Enabled = this.registerForClipboardChangesToolStripMenuItem.Checked;

            // save option to registry so it doesn't have to be set manually everytime the program starts
            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            _log("writing new REREGISTER value");
            if (this.registerForClipboardChangesToolStripMenuItem.Checked) this.baseRKey.SetValue(REREGISTER, 1);
            if (!this.registerForClipboardChangesToolStripMenuItem.Checked) this.baseRKey.SetValue(REREGISTER, 0);
            this.baseRKey.Close();
        }

        private void richTextBox_LinkClicked(object sender, LinkClickedEventArgs e)
        {
            try
            {
                System.Diagnostics.Process.Start(e.LinkText);
                //MessageBox.Show("Link Clicked", "Notice", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            catch // (Exception exc1)
            {
                // see comments below: one exception can be safely ignored, all others can't be handled anyway
                //MessageBox.Show(exc1.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                /*
                // System.ComponentModel.Win32Exception is a known exception that occurs when Firefox is default browser.  
                // It actually opens the browser but STILL throws this exception so we can just ignore it.  If not this exception,
                // then attempt to open the URL in IE instead.
                if (exc1.GetType().ToString() != "System.ComponentModel.Win32Exception")
                {
                    // sometimes throws exception so we have to just ignore
                    // this is a common .NET bug that no one online really has a great reason for so now we just need to try to open
                    // the URL using IE if we can.
                    try
                    {
                        System.Diagnostics.ProcessStartInfo startInfo = new System.Diagnostics.ProcessStartInfo("IExplore.exe", sUrl);
                        System.Diagnostics.Process.Start(startInfo);
                        startInfo = null;
                    }
                    catch (Exception exc2)
                    {
                        // still nothing we can do so just show the error to the user here.
                    }
                }*/
            }
        }

        private void DisableTopMost_Tick(object sender, EventArgs e)
        {
            /*
             * If the "bring to front" option is enabled, also this timer is enabled and should at least disable TopMost.
             * SendToBack is optional, but is supposed to be useful if another, more important window should be in front.
             * 
             * Also, disable the timer, otherwise the window gets sent to back every x seconds.
             */
            this.TopMost = false;
            if (this.bfirstStart)
            {
                this.bfirstStart = false;
            }
            else
            {
                this.SendToBack();
            }
            this.DisableTopMost.Enabled = false;
        }

        private void BringToFrontOption_Click(object sender, EventArgs e)
        {
            // save option to registry so it doesn't have to be set manually everytime the program starts
            this.baseRKey = Registry.CurrentUser.OpenSubKey(SOFTWARE + "\\" + this.ProductName, true);
            _log("writing new BRING TO FRONT value");
            if (this.BringToFrontOption.Checked) this.baseRKey.SetValue(BRING_TO_FRONT, 1);
            if (!this.BringToFrontOption.Checked) this.baseRKey.SetValue(BRING_TO_FRONT, 0);
            this.baseRKey.Close();
        }

        private void Form1_LocationChanged(object sender, EventArgs e)
        {
            this.writeWindowPositionToRegistry();
        }

        private void statusStrip1_DoubleClick(object sender, EventArgs e)
        {
            // toggle hide option that hides all content
            // could be useful if the user wants to copy passwords while being watched
            this.bHideContent = !this.bHideContent;
            // not only toggle option but also toggle display
            this.applyVisibility();
        }

        private void ClipboardChecker_DoWork(object sender, DoWorkEventArgs e)
        {
            /*
             * Check Clipboard and do neccessary calculations here
             */
        }

        private void ClipboardChecker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            /*
             * Do Layouting here (I hope)
             */
        }

        private void showLastException()
        {
            if (this.lastException != null)
            {
                MessageBox.Show("Last caught exception:\r\n\r\n" + this.lastException.ToString());
            }
        }

        private void setShowExceptionMode(ShowExceptionModes s, bool writeToRegistry)
        {
            this.ShowExceptionMode = s;

            this.showExceptionsFullToolStripMenuItem.Checked = false;
            this.showExceptionsStatusBarToolStripMenuItem.Checked = false;
            this.showExceptionsNoneToolStripMenuItem.Checked = false;
            switch (this.ShowExceptionMode)
            {
                case ShowExceptionModes.None:
                    this.showExceptionsNoneToolStripMenuItem.Checked = true;
                    break;
                case ShowExceptionModes.StatusBar:
                    this.showExceptionsStatusBarToolStripMenuItem.Checked = true;
                    break;
                case ShowExceptionModes.ImmediateFull:
                    this.showExceptionsFullToolStripMenuItem.Checked = true;
                    break;
            }
            if (writeToRegistry)
            {
                this.writeShowExceptionModeToRegistry(null);
            }
        }
        
        private void showExceptionsFullToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.setShowExceptionMode(ShowExceptionModes.ImmediateFull, true);
        }

        private void showExceptionsStatusBarToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.setShowExceptionMode(ShowExceptionModes.StatusBar, true);
        }

        private void showExceptionsNoneToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.setShowExceptionMode(ShowExceptionModes.None, true);
        }

        private void typeLabel_Click(object sender, EventArgs e)
        {
            if (this.typeLabel.Text.Contains("Exception"))
            {
                this.showLastException();
            }
        }

        private void notifyIcon1_DoubleClick(object sender, EventArgs e)
        {
            if (this.WindowState == FormWindowState.Minimized || !this.Visible)
            {
                // if minimized or hidden -> show
                this.WindowState = FormWindowState.Normal;
                this.Show();
            }
            else
            {
                // if visible (and not minimized)
                this.Hide();
            };
        }

        private void Form1_SizeChanged(object sender, EventArgs e)
        {
            if (this.WindowState == FormWindowState.Maximized)
            {
                // disallow maximized
                this.WindowState = FormWindowState.Normal;
            }
            else
            {
                if (this.WindowState == FormWindowState.Minimized)
                {
                    // if minimized, restore and hide instead
                    this.WindowState = FormWindowState.Normal;
                    this.Hide();
                }
                else
                {
                    // normal window state -> resize elements
                    this.applySizes();
                };
            };
        }
    }
}
